# coding=utf-8
from __future__ import absolute_import, print_function

import importlib
import os

from suanpan import error


def _importModule(name):
    name = name.replace(os.sep, ".")
    try:
        return importlib.import_module(name)
    except ModuleNotFoundError as e:
        if e.name != name:
            raise e
        return None


def importModule(name):
    module = _importModule(name)
    if module is None:
        raise error.ImportError(f"Module {name} not found")
    return module


def _importVariable(name):
    moduleName, componentName = name.replace(os.sep, ".").rsplit(".", 1)
    module = importModule(moduleName)
    return getattr(module, componentName, None)


def importVariable(name):
    variable = _importVariable(name)
    if variable is None:
        raise error.ImportError(f"Variable {name} not found")
    return variable


def _imports(name):
    name = name.replace(os.sep, ".")
    obj = _importModule(name) or _importVariable(name)
    if obj is None:
        raise error.ImportError(f"{name} not found")
    return obj


def imports(*names):
    objs = tuple(_imports(name) for name in names)
    return objs[0] if len(objs) == 1 else objs



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import g
from suanpan.api.requests import affinity


def listComponents(limit=9999):
    return affinity.post("/component/list", json={"limit": limit})["list"]


def getComponentGraph(componentId):
    return affinity.get(f"/component/graph/{componentId}")["data"]


def updateComponentGraph(componentId, graph):
    return affinity.post(f"/component/graph/{componentId}", json={"graph": graph})


def revertComponentGraph(componentId):
    return affinity.post(f"/component/graph/{componentId}/revert")


def shareComponent(componentId, userId, name):
    return affinity.post(
        "/component/share",
        json={"id": componentId, "targetUserId": userId, "name": name},
    )


def exportModel(portId, name, appId=None, nodeId=None, overwrite=False):
    return affinity.post(
        "/component/export/model",
        json={
            "id": appId or g.appId,
            "nodeId": nodeId or g.nodeId,
            "portId": portId,
            "name": name,
            "overwrite": overwrite,
        },
    )



import suanpan
from suanpan.app import app
from suanpan.app.arguments import String
from suanpan.screenshots import screenshots
from suanpan.utils import image

test_image = image.read("tests/test.png")


@app.afterInit
def clean_screenshots(_):
    screenshots.clean()


@app.param(String(key="param1"))
def test_screenshots(_):
    for _ in range(10):
        screenshots.save(test_image)


if __name__ == "__main__":
    suanpan.run(app)



import numpy as np

import suanpan
from suanpan import path as spath
from suanpan.app import app
from suanpan.app.arguments import Bool, Float, Int, Json, String
from suanpan.log import logger

images = app.modules.enable("images")


@app.afterInit
def afterInit(_):
    spath.copy("tmp/web", app.sio.static(), ignore=True)


@app.title("测试标题")
@app.param(String(key="param1", default="1", label="String Param 1"))
@app.param(Int(key="param2", default=255, label="Int Param 2", step=20))
@app.param(Float(key="param3", default=3.0, label="Float Param 3"))
@app.param(Bool(key="param4", default=True, label="Bool Param 4"))
@app.param(Json(key="param5", default={"test": "json"}, label="Json Param 5"))
def testCall(context):
    args = context.args
    inImage = np.ones((100, 200), dtype=np.uint8) * args.param2
    # 保存处理前图片并触发前端更新
    images.saveDebugImageAndNotify(inImage, itype="input")

    # 自定义机器视觉处理代码
    logger.info("Call Custom CV function")
    outImage = inImage / 2

    # 保存处理后图片并触发前端更新
    images.saveDebugImageAndNotify(outImage, itype="output")


if __name__ == "__main__":
    suanpan.run(app)



import random
import time

import suanpan
from suanpan.app import app


def test_generator():
    for i in range(10000000000):
        yield i
        time.sleep(random.randint(0, 5))


@app.trigger(loop=test_generator)
def test2(_, i):
    print(i)


if __name__ == "__main__":
    suanpan.run(app)



# coding=utf-8
from __future__ import absolute_import, division, print_function

import functools
import math
import os

import oss2
from lostc import collection as lcc
from oss2.models import PartInfo
from oss2.resumable import ResumableDownloadStore, ResumableStore

from suanpan import api, asyncio, g
from suanpan import path as spath
from suanpan import runtime
from suanpan.log import logger
from suanpan.storage import base
from suanpan.utils import pbar as spbar


class Storage(base.Storage):
    LARGE_FILE_SIZE = 1024 * 1024 * 1024  # 1GB
    PREFERRED_SIZE = 500 * 1024 * 1024  # 500MB

    def __init__(
        self,
        ossAccessId=None,
        ossAccessKey=None,
        ossBucketName="suanpan",
        ossEndpoint="http://oss-cn-beijing.aliyuncs.com",
        ossDelimiter="/",
        ossTempStore=base.Storage.DEFAULT_TEMP_DATA_STORE,
        ossGlobalStore=base.Storage.DEFAULT_GLOBAL_DATA_STORE,
        ossDownloadNumThreads=1,
        ossDownloadStoreName=".py-oss-download",
        ossUploadNumThreads=1,
        ossUploadStoreName=".py-oss-upload",
        **kwargs,
    ):
        super().__init__(
            delimiter=ossDelimiter,
            tempStore=ossTempStore,
            globalStore=ossGlobalStore,
            **kwargs,
        )

        self.bucketName = ossBucketName
        self.endpoint = ossEndpoint
        self.refreshAccessKey(accessId=ossAccessId, accessKey=ossAccessKey)

        self.downloadNumThreads = ossDownloadNumThreads
        self.downloadStoreName = ossDownloadStoreName
        self.downloadStore = ResumableDownloadStore(
            self.tempStore, self.downloadStoreName
        )

        self.uploadNumThreads = ossUploadNumThreads
        self.uploadStoreName = ossUploadStoreName
        self.uploadStore = ResumableStore(self.tempStore, self.uploadStoreName)

        self._removeOssLogger()

    def autoRefreshToken(self, func):
        @functools.wraps(func)
        def _dec(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except oss2.exceptions.OssError as e:
                if e.status != 403:
                    raise e
                logger.warning("Oss access denied, refreshing access key.")
                self.refreshAccessKey()
                return func(*args, **kwargs)

        return _dec

    def autoRefreshBucket(self, bucket):
        try:
            bucket.object_exists(f"studio/{g.userId}/config")
            return bucket
        except oss2.exceptions.OssError as e:
            if e.status != 403:
                raise e
            logger.warning("Oss access denied, refreshing access key.")
            self.refreshAccessKey()
            bucket = self.getBucketByName(bucket.bucket_name)
            return bucket

    def refreshAccessKey(self, accessId=None, accessKey=None):
        if accessId and accessKey:
            self.accessId = accessId
            self.accessKey = accessKey
            self.auth = oss2.Auth(self.accessId, self.accessKey)
        else:
            data = api.oss.getToken()
            self.accessId = data["Credentials"]["AccessKeyId"]
            self.accessKey = data["Credentials"]["AccessKeySecret"]
            self.securityToken = data["Credentials"]["SecurityToken"]
            self.auth = oss2.StsAuth(self.accessId, self.accessKey, self.securityToken)

        self.bucket = self.getBucket(self.bucketName)
        return self.accessId, self.accessKey

    def _removeOssLogger(self):
        ossLogger = getattr(oss2, "logger", None)
        if ossLogger:
            self._removeLoggerHandlers(ossLogger)

    def _removeLoggerHandlers(self, _logger):
        for handler in _logger.handlers[:]:
            _logger.removeHandler(handler)
        return _logger

    @runtime.retry(stop_max_attempt_number=3)
    def download(
        self,
        key,
        path=None,
        progress=None,
        bucketOrBucketName=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        downloadFunction = (
            self.downloadFile
            if self.isFile(key, bucketOrBucketName=bucket)
            else self.downloadFolder
        )
        return downloadFunction(
            key,
            path=path,
            progress=progress,
            bucketOrBucketName=bucket,
            ignores=ignores,
            quiet=quiet,
        )

    def downloadFolder(
        self,
        key,
        path=None,
        progress=None,
        delimiter=None,
        bucketOrBucketName=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        delimiter = delimiter or self.delimiter
        path = path or self.getPathInTempStore(key)
        url = self.storageUrl(key, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Downloading folder: {url} -> {path}")

        if self.basename(key) in ignores:
            with spbar.one(
                total=1, desc="Downloading", disable=quiet, progress=progress
            ) as pbar:
                pbar.update(1)
                pbar.set_description("Ignored")
                return path

        downloads = {
            file: self.localPathJoin(path, self.storageRelativePath(file, key))
            for _, _, files in self.walk(
                key, delimiter=delimiter, bucketOrBucketName=bucket
            )
            for file in files
        }

        # Download from oss
        _run = functools.partial(
            self.downloadFile, bucketOrBucketName=bucket, ignores=ignores, quiet=True
        )
        asyncio.starmap(
            _run, downloads.items(), pbar=quiet, desc="Downloading", progress=progress,
        )
        # Remove ignores
        self.removeIgnores(path, ignores=ignores)
        # Remove rest files and folders
        files = (
            os.path.join(root, file)
            for root, _, files in os.walk(path)
            for file in files
        )
        restFiles = [file for file in files if file not in downloads.values()]
        asyncio.map(
            spath.remove,
            restFiles,
            pbar=quiet or not restFiles,
            desc="Removing Rest Files",
        )
        spath.removeEmptyFolders(path)
        if not quiet:
            logger.debug(f"Removed empty folders in: {path}")
        # End
        return path

    def downloadFile(
        self,
        key,
        path=None,
        progress=None,
        bucketOrBucketName=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        path = path or self.getPathInTempStore(key)
        url = self.storageUrl(key, bucketOrBucketName=bucket)
        fileSize = self.getStorageSize(key, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Downloading file: {url} -> {path}")

        with spbar.one(total=fileSize, disable=quiet, progress=progress) as pbar:
            if self.basename(key) in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return path

            fileMd5 = self.getLocalMd5(path)
            objectMd5 = self.getStorageMd5(key, bucketOrBucketName=bucket)
            if self.checkMd5(fileMd5, objectMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return path

            def _percentage(consumed_bytes, total_bytes):
                if total_bytes:
                    pbar.update(consumed_bytes - pbar.n)
                    pbar.set_description("Downloading")

            spath.safeMkdirsForFile(path)
            self.autoRefreshToken(oss2.resumable_download)(
                bucket,
                key,
                path,
                num_threads=self.downloadNumThreads,
                store=self.downloadStore,
                progress_callback=_percentage,
            )

            pbar.set_description("Downloaded")

            return path

    @runtime.retry(stop_max_attempt_number=3)
    def upload(
        self,
        key,
        path=None,
        progress=None,
        bucketOrBucketName=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        uploadFunction = self.uploadFolder if os.path.isdir(path) else self.uploadFile
        return uploadFunction(
            key,
            path=path,
            progress=progress,
            bucketOrBucketName=bucket,
            ignores=ignores,
            quiet=quiet,
        )

    def uploadFolder(
        self,
        key,
        path=None,
        progress=None,
        bucketOrBucketName=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        path = path or self.getPathInTempStore(key)
        url = self.storageUrl(key, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Uploading folder: {path} -> {url}")

        if key in ignores:
            with spbar.one(
                total=1, desc="Uploading", disable=quiet, progress=progress
            ) as pbar:
                pbar.update(1)
                pbar.set_description("Ignored")
                return path

        files = (
            os.path.join(root, file)
            for root, _, files in os.walk(path)
            for file in files
        )
        uploads = {
            file: self.storagePathJoin(key, self.localRelativePath(file, path))
            for file in files
        }

        if not uploads:
            if not quiet:
                logger.warning(f"Uploading empty folder: {path}")
        else:
            # Upload files to oss
            uploadItems = [(key, file) for file, key in uploads.items()]
            _run = functools.partial(
                self.uploadFile, bucketOrBucketName=bucket, ignores=ignores, quiet=True
            )
            asyncio.starmap(
                _run, uploadItems, pbar=quiet, desc="Uploading", progress=progress,
            )

        # Remove rest files
        localFiles = set(self.localRelativePath(file, path) for file in uploads.keys())
        remoteFiles = set(
            self.storageRelativePath(file, key)
            for _, _, files in self.walk(key)
            for file in files
        )
        restFiles = [
            self.storagePathJoin(key, file) for file in remoteFiles - localFiles
        ]
        _run = functools.partial(self.remove, bucketOrBucketName=bucket, quiet=True)
        asyncio.map(
            _run, restFiles, pbar=quiet or not restFiles, desc="Removing Rest Files",
        )
        # End
        return path

    def uploadFile(
        self,
        key,
        path=None,
        progress=None,
        bucketOrBucketName=None,
        ignores=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        ignores = ignores or self.DEFAULT_IGNORE_KEYWORDS
        path = path or self.getPathInTempStore(key)
        url = self.storageUrl(key, bucketOrBucketName=bucket)
        fileSize = self.getLocalSize(path)

        if not quiet:
            logger.info(f"Uploading file: {path} -> {url}")

        with spbar.one(total=fileSize, disable=quiet, progress=progress) as pbar:
            if self.basename(key) in ignores:
                pbar.update(fileSize)
                pbar.set_description("Ignored")
                return path

            fileMd5 = self.getLocalMd5(path)
            objectMd5 = self.getStorageMd5(key, bucketOrBucketName=bucket)
            if self.checkMd5(fileMd5, objectMd5):
                pbar.update(fileSize)
                pbar.set_description("Existed")
                return path

            def _percentage(consumed_bytes, total_bytes):
                if total_bytes:
                    pbar.update(consumed_bytes - pbar.n)
                    pbar.set_description("Uploading")

            self.autoRefreshToken(oss2.resumable_upload)(
                bucket,
                key,
                path,
                num_threads=self.uploadNumThreads,
                store=self.uploadStore,
                progress_callback=_percentage,
                headers={self.CONTENT_MD5: fileMd5},
            )

            pbar.set_description("Uploaded")

            return path

    @runtime.retry(stop_max_attempt_number=3)
    def copy(
        self, name, dist, progress=None, bucketOrBucketName=None, quiet=not g.debug
    ):
        bucket = self.getBucket(bucketOrBucketName)
        copyFunction = (
            self.copyFile
            if self.isFile(name, bucketOrBucketName=bucket)
            else self.copyFolder
        )
        return copyFunction(
            name, dist, progress=progress, bucketOrBucketName=bucket, quiet=quiet
        )

    def copyFolder(
        self,
        src,
        dist,
        progress=None,
        bucketOrBucketName=None,
        delimiter=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        delimiter = delimiter or self.delimiter
        src = self.completePath(src)
        dist = self.completePath(dist)
        srcurl = self.storageUrl(src, bucketOrBucketName=bucket)
        disturl = self.storageUrl(dist, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Copying folder: {srcurl} -> {disturl}")

        copyItems = [
            (file, file.replace(src, dist))
            for _, _, files in self.walk(
                src, delimiter=delimiter, bucketOrBucketName=bucket
            )
            for file in files
        ]
        _run = functools.partial(self.copyFile, bucketOrBucketName=bucket, quiet=True)
        asyncio.starmap(
            _run, copyItems, pbar=quiet, desc="Copying", progress=progress,
        )
        return dist

    def copyFile(
        self, src, dist, progress=None, bucketOrBucketName=None, quiet=not g.debug
    ):
        bucket = self.getBucket(bucketOrBucketName)
        fileSize = self.getStorageSize(src, bucketOrBucketName=bucket)
        copyFunction = (
            self.copyLargeFile
            if fileSize >= self.LARGE_FILE_SIZE
            else self.copySmallFile
        )
        return copyFunction(
            src,
            dist,
            fileSize,
            progress=progress,
            bucketOrBucketName=bucket,
            quiet=quiet,
        )

    def copySmallFile(
        self,
        src,
        dist,
        size,
        progress=None,
        bucketOrBucketName=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        srcurl = self.storageUrl(src, bucketOrBucketName=bucket)
        disturl = self.storageUrl(dist, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Copying file: {srcurl} -> {disturl}")

        with spbar.one(
            total=size, desc="Copying", disable=quiet, progress=progress
        ) as pbar:
            objectMd5 = self.getStorageMd5(src, bucketOrBucketName=bucket)
            distMd5 = self.getStorageMd5(dist, bucketOrBucketName=bucket)
            if self.checkMd5(objectMd5, distMd5):
                pbar.update(size)
                pbar.set_description("Existed")
                return dist

            bucket.copy_object(bucket.bucket_name, src, dist)
            pbar.update(size)
            pbar.set_description("Copied")
            return dist

    def copyLargeFile(
        self,
        src,
        dist,
        size,
        progress=None,
        bucketOrBucketName=None,
        quiet=not g.debug,
    ):
        bucket = self.getBucket(bucketOrBucketName)
        srcurl = self.storageUrl(src, bucketOrBucketName=bucket)
        disturl = self.storageUrl(dist, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Copying file: {srcurl} -> {disturl}")

        with spbar.one(
            total=size, desc="Copying", disable=quiet, progress=progress
        ) as pbar:
            objectMd5 = self.getStorageMd5(src, bucketOrBucketName=bucket)
            distMd5 = self.getStorageMd5(dist, bucketOrBucketName=bucket)
            if self.checkMd5(objectMd5, distMd5):
                pbar.update(size)
                pbar.set_description("Existed")
                return dist

            partSize = self.autoRefreshToken(oss2.determine_part_size)(
                size, preferred_size=self.PREFERRED_SIZE
            )
            uploadId = bucket.init_multipart_upload(dist).upload_id
            parts = math.ceil(size / partSize)
            parts = (
                (i + 1, i * partSize, min((i + 1) * partSize, size))
                for i in range(parts)
            )

            def _copy(part):
                partNumber, byteRange = part[0], part[-2:]
                result = bucket.upload_part_copy(
                    bucket.bucket_name, src, byteRange, dist, uploadId, partNumber,
                )
                pbar.update(byteRange[1] - byteRange[0])
                pbar.set_description("Copying")
                return PartInfo(partNumber, result.etag)

            parts = [_copy(part) for part in parts]
            bucket.complete_multipart_upload(dist, uploadId, parts)
            pbar.set_description("Copied")
            return dist

    @runtime.retry(stop_max_attempt_number=3)
    def remove(
        self,
        key,
        progress=None,
        delimiter=None,
        bucketOrBucketName=None,
        quiet=not g.debug,
    ):
        delimiter = delimiter or self.delimiter
        bucket = self.getBucket(bucketOrBucketName)
        removeFunc = (
            self.removeFile
            if self.isFile(key, bucketOrBucketName=bucket)
            else self.removeFolder
        )
        return removeFunc(
            key,
            progress=progress,
            delimiter=delimiter,
            bucketOrBucketName=bucket,
            quiet=quiet,
        )

    def removeFolder(
        self,
        key,
        progress=None,
        delimiter=None,
        bucketOrBucketName=None,
        quiet=not g.debug,
    ):
        delimiter = delimiter or self.delimiter
        bucket = self.getBucket(bucketOrBucketName)
        key = self.completePath(key)
        url = self.storageUrl(key, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Removing folder: {url}")

        removes = [file for _, _, files in self.walk(key) for file in files]
        _run = functools.partial(
            self.removeFile, delimiter=delimiter, bucketOrBucketName=bucket, quiet=True
        )
        asyncio.map(
            _run,
            removes,
            pbar=quiet or not removes,
            desc="Removing",
            progress=progress,
        )
        return key

    def removeFile(
        self,
        key,
        progress=None,
        delimiter=None,
        bucketOrBucketName=None,
        quiet=not g.debug,
    ):  # pylint: disable=unused-argument
        delimiter = delimiter or self.delimiter
        bucket = self.getBucket(bucketOrBucketName)
        url = self.storageUrl(key, bucketOrBucketName=bucket)

        if not quiet:
            logger.info(f"Removing file: {url}")

        with spbar.one(
            total=1, disable=quiet, desc="Removing", progress=progress,
        ) as pbar:
            bucket.delete_object(key)
            pbar.update(1)
            pbar.set_description("Removed")
        return key

    def walk(self, key, delimiter=None, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        delimiter = delimiter or self.delimiter
        root = self.completePath(key, delimiter=delimiter)
        objects = (
            obj
            for obj in runtime.saferun(self._listAll, default=iter(()))(
                key, delimiter=delimiter, bucketOrBucketName=bucket
            )
            if obj.key != root
        )
        folders, files = lcc.divide(objects, lambda obj: obj.is_prefix())
        folders, files = self._getObjectNames(folders), self._getObjectNames(files)
        yield root, folders, files
        for folder in folders:
            yield from self.walk(folder, delimiter=delimiter, bucketOrBucketName=bucket)

    def _listAll(self, key, delimiter=None, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        delimiter = delimiter or self.delimiter
        prefix = self.completePath(key, delimiter=delimiter)
        return (
            obj
            for obj in self.autoRefreshToken(oss2.ObjectIterator)(
                delimiter=delimiter, prefix=prefix, bucket=bucket
            )
        )

    def listAll(self, key, delimiter=None, bucketOrBucketName=None):
        return (
            obj.key
            for obj in self._listAll(
                key, delimiter=delimiter, bucketOrBucketName=bucketOrBucketName
            )
        )

    def listFolders(self, key, delimiter=None, bucketOrBucketName=None):
        return (
            obj.key
            for obj in self._listAll(
                key, delimiter=delimiter, bucketOrBucketName=bucketOrBucketName
            )
            if obj.is_prefix()
        )

    def listFiles(self, key, delimiter=None, bucketOrBucketName=None):
        return (
            obj.key
            for obj in self._listAll(
                key, delimiter=delimiter, bucketOrBucketName=bucketOrBucketName
            )
            if not obj.is_prefix()
        )

    def isFolder(self, key, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return next(self.listAll(key, bucketOrBucketName=bucket), None)

    def isFile(self, key, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return bucket.object_exists(key)

    def getStorageMd5(self, key, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        try:
            return bucket.head_object(key).headers.get(self.CONTENT_MD5)
        except oss2.exceptions.NotFound:
            return None

    def getStorageSize(self, key, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return bucket.head_object(key).content_length

    def storageUrl(self, key, bucketOrBucketName=None):
        bucket = self.getBucket(bucketOrBucketName)
        return "oss:///" + self.storagePathJoin(bucket.bucket_name, key)

    def getBucket(self, bucketOrBucketName):
        return self.autoRefreshBucket(
            bucketOrBucketName
            if isinstance(bucketOrBucketName, oss2.Bucket)
            else self.getBucketByName(bucketOrBucketName)
        )

    def getBucketByName(self, bucketName=None):
        return (
            oss2.Bucket(self.auth, self.endpoint, bucketName)
            if bucketName
            else self.bucket
        )

    def _getObjectNames(self, objects):
        return (
            [obj.key for obj in objects]
            if isinstance(objects, (tuple, list))
            else objects.key
        )



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.proxy import Proxy


class Graph(Proxy):
    DEFAULT_TYPE = "default"
    MAPPING = {
        "default": "suanpan.graph.base.Graph",
    }


graph = Graph()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import error
from suanpan.arguments.auto import ProxyArg
from suanpan.node import node


class DataFrame(ProxyArg):
    MAPPING = {
        "data": "suanpan.storage.arguments.Csv",
        "table": "suanpan.dw.arguments.Table",
    }

    def __init__(self, key, table, partition, *args, **kwargs):
        port = node.get(table) or node.get(key)
        if not port:
            argsString = f"key={key}, table={table}, partition={partition}"
            raise error.NodeError(f"No such {self.name}: {argsString}")
        kwargs.setdefault(self.TYPE_KEY, port.type)
        kwargs.update(key=key, table=table, partition=partition)
        super().__init__(*args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

import functools


def instancemethod(func):
    @functools.wraps(func)
    def _dec(self, *args, **kwargs):  # pylint: disable=unused-argument
        return func(*args, **kwargs)

    return _dec


def bindmethod(instance, func, name=None):
    name = name or func.__name__
    func = func.__get__(instance, instance.__class__)
    setattr(instance, name, func)
    return instance


def bind(instance, func, name=None):
    return bindmethod(instance, instancemethod(func), name=name)


def onlyonce(func):
    setattr(func, "__onlyonce_called__", False)
    setattr(func, "__onlyonce_return__", None)

    @functools.wraps(func)
    def _dec(*args, **kwargs):
        if not getattr(func, "__onlyonce_called__", False):
            setattr(func, "__onlyonce_called__", True)
            setattr(func, "__onlyonce_return__", func(*args, **kwargs))
        return getattr(func, "__onlyonce_return__")

    return _dec


def lazyproperty(func):
    attrname = "_lazy_" + func.__name__

    @property
    def _lazyproperty(self):
        if not hasattr(self, attrname):
            setattr(self, attrname, func(self))
        return getattr(self, attrname)

    return _lazyproperty



# coding=utf-8
from __future__ import absolute_import, print_function

from contextlib import contextmanager

import pandas as pd

from suanpan import utils
from suanpan.log import logger


class RelationalDataWarehouse(object):
    MAX_SQL_LENGTH = 120
    DEFAULT_DB_NAME = "None"
    DTYPE_SQLTYPE_MAPPING = {
        "int8": "tinyint",
        "int16": "smallint",
        "int32": "int",
        "int64": "bigint",
        "float32": "float",
        "float64": "double",
        "class": "text",
        "datetime": "timestamp",
    }

    def __init__(
        self,
        host,
        port,
        user,
        password,
        database,
        **kwargs,  # pylint: disable=unused-argument
    ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database

    @contextmanager
    def connect(self):
        pass

    def getColumnsNamesFromCursor(self, cursor):  # pylint: disable=unused-argument
        raise NotImplementedError("Method not implemented.")

    @contextmanager
    def cursor(self, connection):
        cursor = connection.cursor()
        yield cursor
        cursor.close()

    def testConnection(self, connection):
        with self.cursor(connection) as cursor:
            self.selectCurrentTimestamp(cursor)
        logger.debug("Database connected")

    def readTable(self, table, partition=None, database=None, columns=None):
        database = self.database if not database else database
        sql = (
            "select {} from {}.{}".format(", ".join(columns), database, table)
            if columns
            else f"select * from {database}.{table}"
        )

        if partition and partition.strip():
            conditions = " and ".join(x.strip() for x in partition.split(","))
            sql = f"{sql} where {conditions}"

        with self.connect() as connection:
            data = pd.read_sql(sql, connection)
        data = data.rename(index=str, columns=lambda x: x.split(".")[-1])

        return data

    def writeTable(self, table, data, database=None, overwrite=True):
        database = database if database else self.database
        columns = self.getColumns(data)
        with self.connect() as connection:
            try:
                cursor = connection.cursor()
                self.useDatabase(cursor, database)
                if overwrite:
                    self.dropTable(cursor, table)
                self.createTable(cursor, table, columns)
                self.insertData(connection, cursor, table, data)
            finally:
                cursor.close()

    def execute(self, sql):
        with self.connect() as connection:
            with self.cursor(connection) as cursor:
                return self._execute(cursor, sql)

    def executeMany(self, sql, data):
        with self.connect() as connection:
            with self.cursor(connection) as cursor:
                cursor.executemany(sql, data.values.tolist())
            connection.commit()

    def createTable(self, cursor, table, columns):
        columnsString = ", ".join([f"{c['name']} {c['type']}" for c in columns])
        self._execute(cursor, f"create table if not exists {table} ({columnsString})")

    def dropTable(self, cursor, table):
        self._execute(cursor, f"drop table if exists {table}")

    def insertData(self, connection, cursor, table, data, per=10000):
        dataLength = len(data)
        logger.debug(f"Insert into tabel {table}: total {dataLength}")
        pieces = dataLength // per + 1
        for i in range(pieces):
            begin = per * i
            end = per * (i + 1)
            values = data[begin:end]
            self.insertPiece(connection, cursor, table, values, begin=begin)

    def insertPiece(self, connection, cursor, table, data, begin=0):
        if not data.empty:
            end = begin + len(data)
            logger.debug(f"[{begin} - {end}] inserting...")
            valuesString = ", ".join(data.apply(self._valueGroup, axis=1).values)
            self._execute(cursor, f"insert into {table} values {valuesString}")
            connection.commit()
            logger.debug(f"[{begin} - {end}] Done.")

    def _valueGroup(self, d):
        groupString = ", ".join([f'"{i}"' for i in d])
        return f"({groupString})"

    def useDatabase(self, cursor, database):
        self._execute(cursor, f"use {database}")

    def selectCurrentTimestamp(self, cursor):
        self._execute(cursor, "select current_timestamp")

    def getColumns(self, data):
        return [
            {"name": name.split(".")[-1], "type": self.toSqlType(dtype)}
            for name, dtype in data.dtypes.items()
        ]

    def toSqlType(self, dtype):
        typeName = self.shortenDatetimeType(dtype.name)
        return self.DTYPE_SQLTYPE_MAPPING.get(typeName, "text")

    def shortenDatetimeType(self, typeName):
        datatimeType = "datetime"
        return datatimeType if typeName.startswith(datatimeType) else typeName

    def _execute(self, cursor, sql):
        try:
            cursor.execute(sql)
            return self._processResults(cursor) if cursor.description else None

        except Exception as e:
            logger.error(utils.shorten(f"SQL: {sql}", self.MAX_SQL_LENGTH))
            raise e

    def _processResults(self, cursor):
        results = cursor.fetchall()
        columnNames = self.getColumnsNamesFromCursor(cursor)
        return pd.DataFrame(results, columns=columnNames)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import path, runtime
from suanpan.arguments import Arg
from suanpan.components import Result
from suanpan.storage import storage
from suanpan.utils import csv, excel, image, json, npy, text


class StorageArg(Arg):
    def getOutputTmpValue(self, *args):
        return storage.delimiter.join(args)


class File(StorageArg):
    FILENAME = "file"
    FILETYPE = None

    def __init__(self, key, **kwargs):
        fileName = kwargs.pop("name", self.FILENAME)
        fileType = kwargs.pop("type", self.FILETYPE)
        self.fileName = f"{fileName}.{fileType.lower()}" if fileType else fileName
        self.objectPrefix = None
        self.objectName = None
        self.filePath = None
        super().__init__(key, **kwargs)

    @property
    def isSet(self):
        return True

    def load(self, args):
        super().load(args)
        if self.value:
            self.objectPrefix = self.value
            self.objectName = storage.storagePathJoin(self.objectPrefix, self.fileName)
            self.filePath = storage.getPathInTempStore(self.objectName)
            self.value = self.filePath
        return self

    def transform(self, value):
        if self.filePath:
            _download = (
                storage.download if self.required else runtime.saferun(storage.download)
            )
            _download(self.objectName, self.filePath)
        return self.filePath

    def clean(self):
        if self.filePath:
            path.remove(self.filePath)
        return self

    def save(self, result):
        filePath = result.value
        storage.upload(self.objectName, filePath)
        if filePath != self.filePath:
            path.remove(filePath)
        self.logSaved(self.objectName)
        return self.objectPrefix


class Folder(StorageArg):
    def __init__(self, key, **kwargs):
        super().__init__(key, **kwargs)
        self.folderName = None
        self.folderPath = None

    @property
    def isSet(self):
        return True

    def load(self, args):
        super().load(args)
        if self.value:
            self.folderName = self.value
            self.folderPath = storage.getPathInTempStore(self.folderName)
            self.value = self.folderPath
        return self

    def transform(self, value):
        if self.folderPath:
            _download = (
                storage.download if self.required else runtime.saferun(storage.download)
            )
            _download(self.folderName, self.folderPath)
        return self.folderPath

    def clean(self):
        if self.folderPath:
            path.empty(self.folderPath)
        return self

    def save(self, result):
        folderPath = result.value
        storage.upload(self.folderName, folderPath)
        if folderPath != self.folderPath:
            path.remove(folderPath)
        self.logSaved(self.folderName)
        return self.folderName


class Data(File):
    FILENAME = "data"


class Json(Data):
    FILETYPE = "json"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = json.load if self.required else runtime.saferun(json.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        json.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Csv(Data):
    FILETYPE = "csv"

    def cleanParams(self, params):
        params = super().cleanParams(params)
        return {k: v for k, v in params.items() if k not in ("table", "partition")}

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = csv.load if self.required else runtime.saferun(csv.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        csv.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Table(Csv):
    pass


class Excel(Data):
    FILETYPE = "xlsx"
    LEGACY_FILETYPE = "xls"

    def __init__(self, key, **kwargs):
        self.fileTypeSet = kwargs.get("type") is not None
        super().__init__(key, **kwargs)

    def load(self, args):
        super().load(args)
        if not self.objectName:
            self.value = self.filePath
            return self

        if not storage.isFile(self.objectName) and not self.fileTypeSet:
            fileType = self.LEGACY_FILETYPE
            self.fileName = "{}.{}".format(self.fileName.split(".")[0], fileType)
            self.objectName = storage.storagePathJoin(self.objectPrefix, self.fileName)
        self.filePath = storage.getPathInTempStore(self.objectName)
        path.safeMkdirsForFile(self.filePath)
        self.value = self.filePath
        return self

    def transform(self, value):
        value = super().transform(value)
        if self.filePath:
            _load = excel.load if self.required else runtime.saferun(excel.load)
            value = _load(self.filePath)
        return value

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        excel.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Npy(Data):
    FILETYPE = "npy"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = npy.load if self.required else runtime.saferun(npy.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        npy.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class String(Data):
    FILETYPE = "txt"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _load = text.load if self.required else runtime.saferun(text.load)
        return _load(filePath)

    def save(self, result):
        path.mkdirs(self.filePath, parent=True)
        text.dump(result.value, self.filePath)
        return super().save(Result.froms(value=self.filePath))


class Visual(String):
    FILENAME = "part-00000"
    FILETYPE = ""


class Model(File):
    FILENAME = "model"


class H5Model(Model):
    FILETYPE = "h5"


class Checkpoint(Model):
    FILETYPE = "ckpt"


class JsonModel(Model):
    FILETYPE = "json"


class Image(File):
    FILENAME = "image"
    FILETYPE = "png"

    def transform(self, value):
        filePath = super().transform(value)
        if not filePath:
            return None
        _read = image.read if self.required else runtime.saferun(image.read)
        return _read(filePath)

    def save(self, result):
        image.save(self.filePath, result.value)
        return super().save(Result.froms(value=self.filePath))


class Screenshots(Image):
    FILENAME = "screenshots"



# coding=utf-8
from __future__ import absolute_import, print_function

import abc
import datetime
import time

from suanpan import interfaces, objects


class Timer(object):
    def __init__(self, delta):
        self.delta = delta
        self.reset()

    def set(self, delta):
        self.delta = delta
        return self

    def reset(self):
        self.current = datetime.datetime.utcnow()
        return self

    @property
    def next(self):
        return self.current + self.delta


class Loop(objects.HasName, interfaces.HasLoopHooks):
    __metaclass__ = abc.ABCMeta

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.options = {"disabled": False, **kwargs}
        self._cache = None

    def __call__(self):
        for x in self.loop():
            if not self.disabled:
                yield x
            if self._cache:
                self._set(**self._cache)

    def beforeLoop(self):
        if self._cache:
            self._set(**self._cache)

    def afterLoop(self):
        if self._cache:
            self._set(**self._cache)

    def before(self, hook):
        self.addBeforeLoopHooks(hook)
        return hook

    def after(self, hook):
        self.addAfterLoopHooks(hook)
        return hook

    @property
    def disabled(self):
        return self.options["disabled"]

    @disabled.setter
    def disabled(self, value):
        self.set(disabled=value)

    def disable(self):
        self.disabled = False
        return self

    def enable(self):
        self.disabled = True
        return self

    def _set(self, **kwargs):
        self.options.update(kwargs)
        self._cache = None
        return self

    def set(self, **kwargs):
        self._cache = kwargs
        return self

    @abc.abstractmethod
    def loop(self):
        pass

    def json(self):
        return {**self.options}


class Interval(Loop):
    def __init__(self, seconds, pre=False):
        super().__init__(seconds=seconds, pre=pre)
        self.timer = Timer(datetime.timedelta(seconds=seconds))
        self.n = 0

    @property
    def seconds(self):
        return self.options["seconds"]

    @seconds.setter
    def seconds(self, value):
        self.set(seconds=value)

    @property
    def pre(self):
        return self.options["pre"]

    @pre.setter
    def pre(self, value):
        self.set(pre=value)

    def _set(self, **kwargs):
        super()._set(**kwargs)
        self.timer.set(datetime.timedelta(seconds=self.seconds))
        return self

    def _loop(self):
        while True:
            self.timer.reset()
            self.callBeforeLoopHooks()
            if self.pre:
                time.sleep(self.seconds)
            yield
            if not self.pre:
                time.sleep(self.seconds)
            self.callAfterLoopHooks()

    def loop(self):
        for i, _ in enumerate(self._loop()):
            self.n = i
            yield

    @property
    def next(self):
        return self.timer.next

    def json(self):
        return {**super().json(), "next": self.next}


class IntervalIndex(Interval):
    def loop(self):
        for _ in super().loop():
            yield self.n



# coding=utf-8
from __future__ import absolute_import, print_function

from contextlib import contextmanager

import psycopg2

from suanpan.dw import base


class DataWarehouse(base.RelationalDataWarehouse):
    def __init__(  # pylint: disable=useless-super-delegation
        self,
        postgresHost,
        postgresPort,
        postgresUsername,
        postgresPassword,
        postgresDatabase,
        **kwargs
    ):
        super().__init__(
            postgresHost,
            postgresPort,
            postgresUsername,
            postgresPassword,
            postgresDatabase,
            **kwargs
        )

    @contextmanager
    def connect(self):
        connection = psycopg2.connect(
            host=self.host,
            port=self.port,
            database=self.database,
            user=self.user,
            password=self.password,
        )

        yield connection
        connection.close()

    def getColumnsNamesFromCursor(self, cursor):
        return [column[0] for column in cursor.description]



# coding=utf-8
from __future__ import absolute_import, print_function

import time
import socketio

from suanpan.api import auth


def sio(*args, **kwargs):
    retryMaxTimes = kwargs.pop("retryMaxTimes", 30)
    retryDelay = kwargs.pop("retryDelay", 0.1)
    ctx = dict(connected=False)
    kwargs["headers"] = {**auth.defaultHeaders(), **kwargs.pop("headers", {})}
    client = socketio.Client()
    client.on("connect", lambda: ctx.update(connected=True))
    client.connect(*args, **kwargs)
    if retryMaxTimes:
        for _ in range(retryMaxTimes):
            if ctx.get("connected", False):
                break
            time.sleep(retryDelay)
        else:
            if not ctx.get("connected", False):
                raise Exception("sio sonnected failed")
    return client
